;;; -*- Mode: Lisp; Package: STELLA; Syntax: COMMON-LISP; Base: 10 -*-

#|--------------------------------------------------------------------------+
 |                                                                          |
 |  COPYRIGHT (C) UNIVERSITY OF SOUTHERN CALIFORNIA, 1997-2000              | 
 |  University of Southern California, Information Sciences Institute       |
 |  4676 Admiralty Way                                                      |
 |  Marina Del Rey, California 90292                                        |
 |                                                                          |
 |  This software was developed under the terms and conditions of Contract  |
 |  No. N00014-94-C-0245 between the Defense Advanced Research Projects     |
 |  Agency and the University of Southern California, Information Sciences  | 
 |  Institute.  Use and distribution of this software is further subject    |
 |  to the provisions of that contract and any other agreements developed   |
 |  between the user of the software and the University of Southern         |
 |  California, Information Sciences Institute.  It is supplied "AS IS",    |
 |  without any warranties of any kind.  It is furnished only on the basis  |
 |  that any party who receives it indemnifies and holds harmless the       |
 |  parties who furnish and originate it against any claims, demands, or    |
 |  liabilities connected with using it, furnishing it to others or         |
 |  providing it to a third party.  THIS NOTICE MUST NOT BE REMOVED FROM    |
 |  THE SOFTWARE, AND IN THE EVENT THAT THE SOFTWARE IS DIVIDED, IT SHOULD  |
 |  BE ATTACHED TO EVERY PART.                                              |
 |                                                                          |
 +--------------------------------------------------------------------------|#

(CL:IN-PACKAGE "STELLA")

;;; Auxiliary variables:

(CL:DEFVAR SGT-PRINT-LOGIC-DESCRIPTION NULL)
(CL:DEFVAR SGT-PRINT-LOGIC-PATTERN-VARIABLE NULL)
(CL:DEFVAR SGT-PRINT-LOGIC-SKOLEM NULL)
(CL:DEFVAR SGT-PRINT-LOGIC-LOGIC-OBJECT NULL)
(CL:DEFVAR SGT-PRINT-LOGIC-PROPOSITION NULL)
(CL:DEFVAR KWD-PRINT-FLAT NULL)
(CL:DEFVAR KWD-PRINT-REALISTIC NULL)
(CL:DEFVAR KWD-PRINT-ORIGINAL NULL)
(CL:DEFVAR KWD-PRINT-DEBUG-LOW NULL)
(CL:DEFVAR KWD-PRINT-DEBUG-HIGH NULL)
(CL:DEFVAR SYM-PRINT-LOGIC-DESCRIPTION NULL)
(CL:DEFVAR SGT-PRINT-STELLA-RELATION NULL)
(CL:DEFVAR SYM-PRINT-LOGIC-DESCRIPTIVE? NULL)
(CL:DEFVAR SYM-PRINT-STELLA-SURROGATE-VALUE-INVERSE NULL)
(CL:DEFVAR SYM-PRINT-LOGIC-STARTUP-PRINT NULL)
(CL:DEFVAR SYM-PRINT-STELLA-METHOD-STARTUP-CLASSNAME NULL)

;;; Forward declarations:

(CL:DECLAIM
 (CL:SPECIAL *STARTUP-TIME-PHASE* *MODULE* *LOADINGREGENERABLEOBJECTS?*
  FALSE-WRAPPER *PROPOSITIONRECORDBINDINGS* *QUERYITERATOR*
  *PRINTREADABLY?* TRUE-TRUTH-VALUE FALSE-TRUTH-VALUE
  DEFAULT-TRUE-TRUTH-VALUE DEFAULT-FALSE-TRUTH-VALUE STANDARD-OUTPUT))

;;; (DEFUN (WRAP-NATIVE-OUTPUT-STREAM OUTPUT-STREAM) ...)

(CL:DEFUN WRAP-NATIVE-OUTPUT-STREAM (NATIVESTREAM)
  (CL:IF (CL:EQ NATIVESTREAM (%NATIVE-STREAM STANDARD-OUTPUT))
   STANDARD-OUTPUT
   (CL:LET* ((SELF-000 (NEW-OUTPUT-STREAM)))
    (CL:SETF (%NATIVE-STREAM SELF-000) NATIVESTREAM)
    (CL:LET* ((VALUE-000 SELF-000)) VALUE-000))))

;;; (DEFUN PRINT-TOP-LEVEL-OBJECT ...)

(CL:DEFUN PRINT-TOP-LEVEL-OBJECT (SELF NATIVESTREAM)
  (CL:LET* ((STREAM (WRAP-NATIVE-OUTPUT-STREAM NATIVESTREAM)))
   (CL:WHEN (DELETED? SELF)
    (%%PRINT-STREAM (%NATIVE-STREAM STREAM) "#DELETED#"))
   (CL:LET* ((TEST-VALUE-000 (SAFE-PRIMARY-TYPE SELF)))
    (CL:COND
     ((SUBTYPE-OF? TEST-VALUE-000 SGT-PRINT-LOGIC-DESCRIPTION)
      (CL:PROGN (PRINT-DESCRIPTION SELF STREAM TRUE)))
     ((SUBTYPE-OF? TEST-VALUE-000 SGT-PRINT-LOGIC-PATTERN-VARIABLE)
      (CL:PROGN (%%PRINT-STREAM (%NATIVE-STREAM STREAM) "|V|")
       (PRINT-VARIABLE SELF STREAM)))
     ((SUBTYPE-OF? TEST-VALUE-000 SGT-PRINT-LOGIC-SKOLEM)
      (CL:PROGN (PRINT-SKOLEM SELF STREAM TRUE)))
     ((SUBTYPE-OF? TEST-VALUE-000 SGT-PRINT-LOGIC-LOGIC-OBJECT)
      (CL:PROGN (PRINT-LOGIC-OBJECT SELF STREAM TRUE)))
     ((SUBTYPE-OF? TEST-VALUE-000 SGT-PRINT-LOGIC-PROPOSITION)
      (CL:PROGN (PRINT-PROPOSITION SELF STREAM TRUE)))
     (CL:T (%%PRINT-STREAM (%NATIVE-STREAM STREAM) SELF))))))

;;; (DEFSPECIAL *PRINTMODE* ...)

(CL:DEFVAR *PRINTMODE* NULL
  "Controls the kind of detail that gets printed about
individual objects.  Values are :ORIGINAL, :REALISTIC, :FLAT, :DEBUG-LOW, and
 :DEBUG-HIGH.")

;;; (DEFSPECIAL *PRINTFUNCTIONSASRELATIONS?* ...)

(CL:DEFVAR *PRINTFUNCTIONSASRELATIONS?* FALSE
  "Controls whether functions are printed as relations (F i v) or
as functions (= (F i) v).")

;;; (DEFUN PRINT-OBJECT-NAME ...)

(CL:DEFUN PRINT-OBJECT-NAME (SELF STREAM)
  (CL:LET* ((TEST-VALUE-000 (SAFE-PRIMARY-TYPE SELF)))
   (CL:COND
    ((SUBTYPE-OF? TEST-VALUE-000 SGT-PRINT-LOGIC-SKOLEM)
     (CL:PROGN
      (%%PRINT-STREAM (%NATIVE-STREAM STREAM)
       (STRING-DOWNCASE (GET-SKOLEM-PRINT-NAME SELF)))))
    ((SUBTYPE-OF? TEST-VALUE-000 SGT-PRINT-LOGIC-LOGIC-OBJECT)
     (CL:PROGN (PRINT-LOGIC-OBJECT SELF STREAM FALSE)))
    ((SUBTYPE-OF-INTEGER? TEST-VALUE-000)
     (CL:PROGN
      (%%PRINT-STREAM (%NATIVE-STREAM STREAM) (%WRAPPER-VALUE SELF))))
    ((SUBTYPE-OF-FLOAT? TEST-VALUE-000)
     (CL:PROGN
      (%%PRINT-STREAM (%NATIVE-STREAM STREAM) (%WRAPPER-VALUE SELF))))
    ((SUBTYPE-OF-STRING? TEST-VALUE-000)
     (CL:PROGN
      (%%PRINT-STREAM (%NATIVE-STREAM STREAM) (%WRAPPER-VALUE SELF))))
    (CL:T (%%PRINT-STREAM (%NATIVE-STREAM STREAM) SELF)))))

;;; (DEFUN PRINT-LOGIC-OBJECT ...)

(CL:DEFUN PRINT-LOGIC-OBJECT (SELF STREAM TOPLEVEL?)
  (CL:LET*
   ((PREFIX
     (CL:IF (CL:AND TOPLEVEL? (CL:NOT *PRINTREADABLY?*)) "|i|" "")))
   (CL:DECLARE (CL:TYPE CL:SIMPLE-STRING PREFIX))
   (CL:COND
    ((CL:NOT (CL:EQ (%SURROGATE-VALUE-INVERSE SELF) NULL))
     (CL:WHEN
      (CL:NOT
       (CL:EQ (%SURROGATE-VALUE (%SURROGATE-VALUE-INVERSE SELF)) SELF))
      (CL:SETQ PREFIX "|BAD surrogate|"))
     (%%PRINT-STREAM (%NATIVE-STREAM STREAM) PREFIX
      (%SYMBOL-NAME (%SURROGATE-VALUE-INVERSE SELF))))
    ((CL:EQ SELF TRUE-TRUTH-VALUE)
     (%%PRINT-STREAM (%NATIVE-STREAM STREAM) "TRUE-TV"))
    ((CL:EQ SELF FALSE-TRUTH-VALUE)
     (%%PRINT-STREAM (%NATIVE-STREAM STREAM) "FALSE-TV"))
    ((CL:EQ SELF DEFAULT-TRUE-TRUTH-VALUE)
     (%%PRINT-STREAM (%NATIVE-STREAM STREAM) "DEFAULT-TRUE-TV"))
    ((CL:EQ SELF DEFAULT-FALSE-TRUTH-VALUE)
     (%%PRINT-STREAM (%NATIVE-STREAM STREAM) "DEFAULT-FALSE-TV"))
    (CL:T
     (%%PRINT-STREAM (%NATIVE-STREAM STREAM) PREFIX "_"
      (%SYMBOL-NAME (LOGICAL-TYPE SELF)))))))

;;; (DEFCONSTANT VARIABLE-BINDING-SEPARATOR ...)

(CL:PROGN (CL:DEFVAR VARIABLE-BINDING-SEPARATOR "/") (CL:DECLAIM (CL:TYPE CL:SIMPLE-STRING VARIABLE-BINDING-SEPARATOR)))

;;; (DEFCONSTANT VALUE-OF-SEPARATOR ...)

(CL:PROGN (CL:DEFVAR VALUE-OF-SEPARATOR "//") (CL:DECLAIM (CL:TYPE CL:SIMPLE-STRING VALUE-OF-SEPARATOR)))

;;; (DEFSPECIAL *SKOLEMNAMEMAPPINGTABLE* ...)

(CL:DEFVAR *SKOLEMNAMEMAPPINGTABLE* NULL
  "If set, maps some skolems to names other than
their own.")

;;; (DEFUN (GET-SKOLEM-PRINT-NAME STRING) ...)

(CL:DEFUN GET-SKOLEM-PRINT-NAME (SELF)
  (CL:WHEN (CL:EQ (%SKOLEM-NAME SELF) NULL)
   (CL:RETURN-FROM GET-SKOLEM-PRINT-NAME ""))
  (CL:IF (CL:NOT (CL:EQ *SKOLEMNAMEMAPPINGTABLE* NULL))
   (CL:LET* ((SUBSTITUTESKOLEM (LOOKUP *SKOLEMNAMEMAPPINGTABLE* SELF)))
    (CL:IF (CL:NOT (CL:EQ SUBSTITUTESKOLEM NULL))
     (CL:IF (CL:EQ (%SKOLEM-NAME SUBSTITUTESKOLEM) NULL) ""
      (%SYMBOL-NAME (%SKOLEM-NAME SUBSTITUTESKOLEM)))
     (%SYMBOL-NAME (%SKOLEM-NAME SELF))))
   (%SYMBOL-NAME (%SKOLEM-NAME SELF))))

;;; (DEFUN PRINT-VALUE-OF-CHAIN ...)

(CL:DEFUN PRINT-VALUE-OF-CHAIN (SELF STREAM CHAINEND)
  (CL:WHEN (CL:EQ CHAINEND NULL) (CL:SETQ CHAINEND SELF))
  (CL:COND
   ((CL:OR (CL:EQ *PRINTMODE* KWD-PRINT-REALISTIC)
     (CL:EQ *PRINTMODE* KWD-PRINT-ORIGINAL))
    (CL:COND
     ((SUBTYPE-OF? (SAFE-PRIMARY-TYPE SELF)
       SGT-PRINT-LOGIC-PATTERN-VARIABLE)
      (CL:PROGN
       (CL:LET* ((*PRINTMODE* KWD-PRINT-FLAT))
        (CL:DECLARE (CL:SPECIAL *PRINTMODE*))
        (PRINT-VALUE-OF-CHAIN SELF STREAM CHAINEND))))
     (CL:T (PRINT-OBJECT-NAME CHAINEND STREAM))))
   ((CL:OR (CL:EQ *PRINTMODE* KWD-PRINT-FLAT)
     (CL:EQ *PRINTMODE* KWD-PRINT-DEBUG-LOW))
    (%%PRINT-STREAM (%NATIVE-STREAM STREAM)
     (STRING-DOWNCASE (GET-SKOLEM-PRINT-NAME SELF)))
    (CL:WHEN (CL:EQ SELF CHAINEND) (CL:RETURN-FROM PRINT-VALUE-OF-CHAIN))
    (%%PRINT-STREAM (%NATIVE-STREAM STREAM) VALUE-OF-SEPARATOR)
    (PRINT-OBJECT-NAME CHAINEND STREAM))
   ((CL:EQ *PRINTMODE* KWD-PRINT-DEBUG-HIGH)
    (%%PRINT-STREAM (%NATIVE-STREAM STREAM) (GET-SKOLEM-PRINT-NAME SELF))
    (CL:WHEN (CL:EQ SELF CHAINEND) (CL:RETURN-FROM PRINT-VALUE-OF-CHAIN))
    (CL:LET* ((CURSOR SELF) (NEXTVALUE NULL))
     (CL:LOOP
      (CL:LET* ((TEST-VALUE-000 (SAFE-PRIMARY-TYPE CURSOR)))
       (CL:COND
        ((SUBTYPE-OF? TEST-VALUE-000 SGT-PRINT-LOGIC-SKOLEM)
         (CL:PROGN
          (CL:SETQ NEXTVALUE
           (ACCESS-IN-CONTEXT (%VARIABLE-VALUE CURSOR)
            (%HOME-CONTEXT CURSOR) FALSE))))
        ((SUBTYPE-OF? TEST-VALUE-000 SGT-PRINT-LOGIC-LOGIC-OBJECT)
         (CL:PROGN (CL:SETQ NEXTVALUE CURSOR)))
        ((SUBTYPE-OF-SURROGATE? TEST-VALUE-000)
         (CL:PROGN (CL:SETQ NEXTVALUE (%SURROGATE-VALUE CURSOR))))
        ((SUBTYPE-OF? TEST-VALUE-000 SGT-PRINT-STELLA-RELATION)
         (CL:PROGN
          (CL:SETQ NEXTVALUE
           (DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS CURSOR)
            SYM-PRINT-LOGIC-DESCRIPTION NULL))))
        (CL:T
         (CL:LET* ((STREAM-000 (NEW-OUTPUT-STRING-STREAM)))
          (%%PRINT-STREAM (%NATIVE-STREAM STREAM-000) "`" TEST-VALUE-000
           "' is not a valid case option")
          (CL:ERROR
           (NEW-STELLA-EXCEPTION (THE-STRING-READER STREAM-000)))))))
      (%%PRINT-STREAM (%NATIVE-STREAM STREAM) VALUE-OF-SEPARATOR)
      (PRINT-OBJECT-NAME NEXTVALUE STREAM)
      (CL:WHEN (EQL? NEXTVALUE CHAINEND) (CL:RETURN))
      (CL:SETQ CURSOR NEXTVALUE))))
   (CL:T
    (CL:LET* ((STREAM-001 (NEW-OUTPUT-STRING-STREAM)))
     (%%PRINT-STREAM (%NATIVE-STREAM STREAM-001) "`" *PRINTMODE*
      "' is not a valid case option")
     (CL:ERROR (NEW-STELLA-EXCEPTION (THE-STRING-READER STREAM-001)))))))

;;; (DEFUN PRINT-VALUE ...)

(CL:DEFUN PRINT-VALUE (SELF STREAM)
  (CL:LET* ((TEST-VALUE-000 (SAFE-PRIMARY-TYPE SELF)))
   (CL:COND
    ((SUBTYPE-OF-INTEGER? TEST-VALUE-000)
     (CL:PROGN
      (%%PRINT-STREAM (%NATIVE-STREAM STREAM) (%WRAPPER-VALUE SELF))))
    ((SUBTYPE-OF-FLOAT? TEST-VALUE-000)
     (CL:PROGN
      (%%PRINT-STREAM (%NATIVE-STREAM STREAM) (%WRAPPER-VALUE SELF))))
    ((SUBTYPE-OF-STRING? TEST-VALUE-000)
     (CL:PROGN
      (%%PRINT-STREAM (%NATIVE-STREAM STREAM) (%WRAPPER-VALUE SELF))))
    ((SUBTYPE-OF-CHARACTER? TEST-VALUE-000)
     (CL:PROGN
      (%%PRINT-STREAM (%NATIVE-STREAM STREAM) (%WRAPPER-VALUE SELF))))
    (CL:T (%%PRINT-STREAM (%NATIVE-STREAM STREAM) SELF)))))

;;; (DEFUN PRINT-SKOLEM ...)

(CL:DEFUN PRINT-SKOLEM (SELF STREAM TOPLEVEL?)
  (CL:LET* ((PREFIX "|SK|"))
   (CL:DECLARE (CL:TYPE CL:SIMPLE-STRING PREFIX))
   (CL:COND
    ((CL:OR (CL:EQ *PRINTMODE* KWD-PRINT-ORIGINAL)
      (CL:EQ *PRINTMODE* KWD-PRINT-REALISTIC))
     (CL:SETQ PREFIX "")
     (CL:WHEN (CL:NOT (CL:EQ (NATIVE-VALUE-OF SELF) NULL))
      (PRINT-VALUE (NATIVE-VALUE-OF SELF) STREAM)
      (CL:RETURN-FROM PRINT-SKOLEM)))
    (CL:T))
   (CL:WHEN
    (CL:AND (CL:NOT (CL:EQ (%DEFINING-PROPOSITION SELF) NULL))
     (CL:EQ
      (ACCESS-IN-CONTEXT (%VARIABLE-VALUE SELF) (%HOME-CONTEXT SELF)
       FALSE)
      NULL))
    (%%PRINT-STREAM (%NATIVE-STREAM STREAM) PREFIX)
    (CL:LET* ((*PRINTLOGICALFORMSTREAM* STREAM) (*INDENTCOUNTER* 20))
     (CL:DECLARE (CL:SPECIAL *PRINTLOGICALFORMSTREAM* *INDENTCOUNTER*))
     (CL:DECLARE (CL:TYPE CL:FIXNUM *INDENTCOUNTER*))
     (PRINT-KIF-FUNCTION-EXPRESSION (%DEFINING-PROPOSITION SELF)))
    (CL:RETURN-FROM PRINT-SKOLEM))
   (PRINT-VALUE-OF-CHAIN SELF STREAM (INNERMOST-OF SELF))))

;;; (DEFUN (BOUND-TO-IN-FRAME OBJECT) ...)

(CL:DEFUN BOUND-TO-IN-FRAME (SELF FRAME)
  (CL:IF
   (CL:NOT
    (CL:EQ (%VARIABLE-BINDINGS (OPERATIVE-PATTERN-RECORD FRAME)) NULL))
   (CL:AREF
    (CL:THE CL:SIMPLE-VECTOR
     (%THE-ARRAY (%VARIABLE-BINDINGS (OPERATIVE-PATTERN-RECORD FRAME))))
    (CL:THE CL:FIXNUM (%BOUND-TO-OFFSET SELF)))
   NULL))

;;; (DEFSPECIAL *PRINTINFRAME* ...)

(CL:DEFVAR *PRINTINFRAME* NULL
  "If set, controls diagnostic printing by making
variable bindings appear relative to the frame '*printInFrame*'.")

;;; (DEFUN PRINT-VARIABLE ...)

(CL:DEFUN PRINT-VARIABLE (SELF STREAM)
  (CL:WHEN
   (CL:AND (CL:NOT (CL:EQ *PROPOSITIONRECORDBINDINGS* NULL))
    (CL:NOT (CL:EQ (LOOKUP *PROPOSITIONRECORDBINDINGS* SELF) NULL)))
   (PRETTY-PRINT-LOGICAL-FORM (LOOKUP *PROPOSITIONRECORDBINDINGS* SELF)
    STREAM)
   (CL:RETURN-FROM PRINT-VARIABLE))
  (CL:WHEN
   (CL:AND (DEFINED? (%BOUND-TO-OFFSET SELF))
    (CL:NOT (CL:EQ *QUERYITERATOR* NULL))
    (CL:NOT
     (CL:EQ (%VARIABLE-BINDINGS (%CURRENT-PATTERN-RECORD *QUERYITERATOR*))
      NULL))
    (CL:< (%BOUND-TO-OFFSET SELF)
     (LENGTH
      (%VARIABLE-BINDINGS (%CURRENT-PATTERN-RECORD *QUERYITERATOR*)))))
   (CL:LET*
    ((VALUE
      (CL:IF (CL:NOT (CL:EQ *PRINTINFRAME* NULL))
       (BOUND-TO-IN-FRAME SELF *PRINTINFRAME*)
       (CL:AREF
        (CL:THE CL:SIMPLE-VECTOR
         (%THE-ARRAY
          (%VARIABLE-BINDINGS (%CURRENT-PATTERN-RECORD *QUERYITERATOR*))))
        (CL:THE CL:FIXNUM (%BOUND-TO-OFFSET SELF))))))
    (CL:WHEN (CL:NOT (CL:EQ VALUE NULL))
     (%%PRINT-STREAM (%NATIVE-STREAM STREAM)
      (STRING-DOWNCASE (GET-SKOLEM-PRINT-NAME SELF))
      VARIABLE-BINDING-SEPARATOR)
     (CL:IF (ISA? VALUE SGT-PRINT-LOGIC-PATTERN-VARIABLE)
      (%%PRINT-STREAM (%NATIVE-STREAM STREAM)
       (GET-SKOLEM-PRINT-NAME VALUE))
      (PRINT-TOP-LEVEL-OBJECT VALUE (%NATIVE-STREAM STREAM)))
     (CL:RETURN-FROM PRINT-VARIABLE))))
  (PRINT-VALUE-OF-CHAIN SELF STREAM (INNERMOST-OF SELF)))

;;; (DEFUN PRINT-QUANTIFIED-VARIABLE ...)

(CL:DEFUN PRINT-QUANTIFIED-VARIABLE (SELF STREAM)
  (PRINT-VARIABLE SELF STREAM))

;;; (DEFUN PRINT-PROPOSITION ...)

(CL:DEFUN PRINT-PROPOSITION (SELF STREAM TOPLEVEL?)
  (CL:LET* ((PREFIX STELLA::NULL-STRING))
   (CL:DECLARE (CL:TYPE CL:SIMPLE-STRING PREFIX))
   (CL:WHEN (DELETED? SELF)
    (%%PRINT-STREAM (%NATIVE-STREAM STREAM) "|dElEtEd-proposition|")
    (CL:RETURN-FROM PRINT-PROPOSITION))
   (CL:WHEN (CL:AND TOPLEVEL? (CL:NOT *PRINTREADABLY?*))
    (CL:COND
     ((CL:OR (DEFAULT-TRUE? SELF) (DEFAULT-FALSE? SELF))
      (CL:SETQ PREFIX
       (CL:IF
        (%WRAPPER-VALUE
         (DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS SELF)
          SYM-PRINT-LOGIC-DESCRIPTIVE? FALSE-WRAPPER))
        "|pd|" "|p|")))
     ((CL:OR (TRUE-PROPOSITION? SELF) (FALSE? SELF))
      (CL:SETQ PREFIX
       (CL:IF
        (%WRAPPER-VALUE
         (DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS SELF)
          SYM-PRINT-LOGIC-DESCRIPTIVE? FALSE-WRAPPER))
        "|Pd|" "|P|")))
     (CL:T
      (CL:SETQ PREFIX
       (CL:IF
        (%WRAPPER-VALUE
         (DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS SELF)
          SYM-PRINT-LOGIC-DESCRIPTIVE? FALSE-WRAPPER))
        "|Pd?|" "|P?|")))))
   (CL:WHEN (COERCE-WRAPPED-BOOLEAN-TO-BOOLEAN (UNFASTENED? SELF))
    (CL:SETQ PREFIX
     (CL:IF
      (%WRAPPER-VALUE
       (DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS SELF)
        SYM-PRINT-LOGIC-DESCRIPTIVE? FALSE-WRAPPER))
      "|uPd|" "|uP|")))
   (CL:WHEN
    (CL:AND
     (CL:NOT
      (CL:EQ
       (DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS SELF)
        SYM-PRINT-STELLA-SURROGATE-VALUE-INVERSE NULL)
       NULL))
     (CL:NOT
      (CL:EQ
       (%SURROGATE-VALUE
        (DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS SELF)
         SYM-PRINT-STELLA-SURROGATE-VALUE-INVERSE NULL))
       SELF)))
    (CL:SETQ PREFIX "|BAD surrogate|"))
   (CL:WHEN (CL:NOT (CL:EQ PREFIX STELLA::NULL-STRING))
    (%%PRINT-STREAM (%NATIVE-STREAM STREAM) PREFIX)))
  (PRINT-LOGICAL-FORM SELF STREAM))

;;; (DEFGLOBAL *DEBUG-MODE?* ...)

(CL:DEFVAR *DEBUG-MODE?* FALSE
  "If set to TRUE, simplify printing so that it
doesn't break during a debugging session.")

;;; (DEFUN PRINT-DESCRIPTION ...)

(CL:DEFUN PRINT-DESCRIPTION (SELF STREAM TOPLEVEL?)
  (CL:WHEN (DELETED? SELF)
   (%%PRINT-STREAM (%NATIVE-STREAM STREAM) "|dElEtEd-description|")
   (CL:WHEN (CL:NOT (CL:EQ (%SURROGATE-VALUE-INVERSE SELF) NULL))
    (%%PRINT-STREAM (%NATIVE-STREAM STREAM)
     (%SURROGATE-VALUE-INVERSE SELF)))
   (CL:RETURN-FROM PRINT-DESCRIPTION))
  (CL:COND
   ((CL:OR *LOADINGREGENERABLEOBJECTS?*
     (COERCE-WRAPPED-BOOLEAN-TO-BOOLEAN (BAD? SELF)) *DEBUG-MODE?*)
    (%%PRINT-STREAM (%NATIVE-STREAM STREAM) "|d|"))
   ((CL:AND TOPLEVEL? (CL:NOT *PRINTREADABLY?*))
    (CL:COND
     ((CLASS? SELF) (%%PRINT-STREAM (%NATIVE-STREAM STREAM) "|c|"))
     ((FUNCTION? SELF) (%%PRINT-STREAM (%NATIVE-STREAM STREAM) "|f|"))
     ((RELATION? SELF) (%%PRINT-STREAM (%NATIVE-STREAM STREAM) "|r|"))
     (CL:T (%%PRINT-STREAM (%NATIVE-STREAM STREAM) "|d|")))))
  (PRINT-LOGICAL-FORM SELF STREAM))

(CL:DEFUN STARTUP-PRINT ()
  (CL:LET*
   ((*MODULE* (GET-STELLA-MODULE "/LOGIC" (> *STARTUP-TIME-PHASE* 1)))
    (*CONTEXT* *MODULE*))
   (CL:DECLARE (CL:SPECIAL *MODULE* *CONTEXT*))
   (CL:WHEN (CURRENT-STARTUP-TIME-PHASE? 2)
    (CL:SETQ SGT-PRINT-LOGIC-DESCRIPTION
     (INTERN-RIGID-SYMBOL-WRT-MODULE "DESCRIPTION" NULL 1))
    (CL:SETQ SGT-PRINT-LOGIC-PATTERN-VARIABLE
     (INTERN-RIGID-SYMBOL-WRT-MODULE "PATTERN-VARIABLE" NULL 1))
    (CL:SETQ SGT-PRINT-LOGIC-SKOLEM
     (INTERN-RIGID-SYMBOL-WRT-MODULE "SKOLEM" NULL 1))
    (CL:SETQ SGT-PRINT-LOGIC-LOGIC-OBJECT
     (INTERN-RIGID-SYMBOL-WRT-MODULE "LOGIC-OBJECT" NULL 1))
    (CL:SETQ SGT-PRINT-LOGIC-PROPOSITION
     (INTERN-RIGID-SYMBOL-WRT-MODULE "PROPOSITION" NULL 1))
    (CL:SETQ KWD-PRINT-FLAT
     (INTERN-RIGID-SYMBOL-WRT-MODULE "FLAT" NULL 2))
    (CL:SETQ KWD-PRINT-REALISTIC
     (INTERN-RIGID-SYMBOL-WRT-MODULE "REALISTIC" NULL 2))
    (CL:SETQ KWD-PRINT-ORIGINAL
     (INTERN-RIGID-SYMBOL-WRT-MODULE "ORIGINAL" NULL 2))
    (CL:SETQ KWD-PRINT-DEBUG-LOW
     (INTERN-RIGID-SYMBOL-WRT-MODULE "DEBUG-LOW" NULL 2))
    (CL:SETQ KWD-PRINT-DEBUG-HIGH
     (INTERN-RIGID-SYMBOL-WRT-MODULE "DEBUG-HIGH" NULL 2))
    (CL:SETQ SYM-PRINT-LOGIC-DESCRIPTION
     (INTERN-RIGID-SYMBOL-WRT-MODULE "DESCRIPTION" NULL 0))
    (CL:SETQ SGT-PRINT-STELLA-RELATION
     (INTERN-RIGID-SYMBOL-WRT-MODULE "RELATION"
      (GET-STELLA-MODULE "/STELLA" TRUE) 1))
    (CL:SETQ SYM-PRINT-LOGIC-DESCRIPTIVE?
     (INTERN-RIGID-SYMBOL-WRT-MODULE "DESCRIPTIVE?" NULL 0))
    (CL:SETQ SYM-PRINT-STELLA-SURROGATE-VALUE-INVERSE
     (INTERN-RIGID-SYMBOL-WRT-MODULE "SURROGATE-VALUE-INVERSE"
      (GET-STELLA-MODULE "/STELLA" TRUE) 0))
    (CL:SETQ SYM-PRINT-LOGIC-STARTUP-PRINT
     (INTERN-RIGID-SYMBOL-WRT-MODULE "STARTUP-PRINT" NULL 0))
    (CL:SETQ SYM-PRINT-STELLA-METHOD-STARTUP-CLASSNAME
     (INTERN-RIGID-SYMBOL-WRT-MODULE "METHOD-STARTUP-CLASSNAME"
      (GET-STELLA-MODULE "/STELLA" TRUE) 0)))
   (CL:WHEN (CURRENT-STARTUP-TIME-PHASE? 4)
    (CL:SETQ *PRINTMODE* KWD-PRINT-FLAT))
   (CL:WHEN (CURRENT-STARTUP-TIME-PHASE? 6) (FINALIZE-CLASSES))
   (CL:WHEN (CURRENT-STARTUP-TIME-PHASE? 7)
    (DEFINE-FUNCTION-OBJECT "WRAP-NATIVE-OUTPUT-STREAM"
     "(DEFUN (WRAP-NATIVE-OUTPUT-STREAM OUTPUT-STREAM) ((NATIVESTREAM NATIVE-OUTPUT-STREAM)))"
     (CL:FUNCTION WRAP-NATIVE-OUTPUT-STREAM) NULL)
    (DEFINE-FUNCTION-OBJECT "PRINT-TOP-LEVEL-OBJECT"
     "(DEFUN PRINT-TOP-LEVEL-OBJECT ((SELF OBJECT) (NATIVESTREAM NATIVE-OUTPUT-STREAM)))"
     (CL:FUNCTION PRINT-TOP-LEVEL-OBJECT) NULL)
    (DEFINE-FUNCTION-OBJECT "PRINT-OBJECT-NAME"
     "(DEFUN PRINT-OBJECT-NAME ((SELF OBJECT) (STREAM OUTPUT-STREAM)))"
     (CL:FUNCTION PRINT-OBJECT-NAME) NULL)
    (DEFINE-FUNCTION-OBJECT "PRINT-LOGIC-OBJECT"
     "(DEFUN PRINT-LOGIC-OBJECT ((SELF LOGIC-OBJECT) (STREAM OUTPUT-STREAM) (TOPLEVEL? BOOLEAN)))"
     (CL:FUNCTION PRINT-LOGIC-OBJECT) NULL)
    (DEFINE-FUNCTION-OBJECT "GET-SKOLEM-PRINT-NAME"
     "(DEFUN (GET-SKOLEM-PRINT-NAME STRING) ((SELF SKOLEM)))"
     (CL:FUNCTION GET-SKOLEM-PRINT-NAME) NULL)
    (DEFINE-FUNCTION-OBJECT "PRINT-VALUE-OF-CHAIN"
     "(DEFUN PRINT-VALUE-OF-CHAIN ((SELF SKOLEM) (STREAM OUTPUT-STREAM) (CHAINEND OBJECT)))"
     (CL:FUNCTION PRINT-VALUE-OF-CHAIN) NULL)
    (DEFINE-FUNCTION-OBJECT "PRINT-VALUE"
     "(DEFUN PRINT-VALUE ((SELF OBJECT) (STREAM OUTPUT-STREAM)))"
     (CL:FUNCTION PRINT-VALUE) NULL)
    (DEFINE-FUNCTION-OBJECT "PRINT-SKOLEM"
     "(DEFUN PRINT-SKOLEM ((SELF SKOLEM) (STREAM OUTPUT-STREAM) (TOPLEVEL? BOOLEAN)))"
     (CL:FUNCTION PRINT-SKOLEM) NULL)
    (DEFINE-FUNCTION-OBJECT "BOUND-TO-IN-FRAME"
     "(DEFUN (BOUND-TO-IN-FRAME OBJECT) ((SELF PATTERN-VARIABLE) (FRAME CONTROL-FRAME)))"
     (CL:FUNCTION BOUND-TO-IN-FRAME) NULL)
    (DEFINE-FUNCTION-OBJECT "PRINT-VARIABLE"
     "(DEFUN PRINT-VARIABLE ((SELF PATTERN-VARIABLE) (STREAM OUTPUT-STREAM)))"
     (CL:FUNCTION PRINT-VARIABLE) NULL)
    (DEFINE-FUNCTION-OBJECT "PRINT-QUANTIFIED-VARIABLE"
     "(DEFUN PRINT-QUANTIFIED-VARIABLE ((SELF PATTERN-VARIABLE) (STREAM OUTPUT-STREAM)))"
     (CL:FUNCTION PRINT-QUANTIFIED-VARIABLE) NULL)
    (DEFINE-FUNCTION-OBJECT "PRINT-PROPOSITION"
     "(DEFUN PRINT-PROPOSITION ((SELF PROPOSITION) (STREAM OUTPUT-STREAM) (TOPLEVEL? BOOLEAN)))"
     (CL:FUNCTION PRINT-PROPOSITION) NULL)
    (DEFINE-FUNCTION-OBJECT "PRINT-DESCRIPTION"
     "(DEFUN PRINT-DESCRIPTION ((SELF DESCRIPTION) (STREAM OUTPUT-STREAM) (TOPLEVEL? BOOLEAN)))"
     (CL:FUNCTION PRINT-DESCRIPTION) NULL)
    (DEFINE-FUNCTION-OBJECT "STARTUP-PRINT"
     "(DEFUN STARTUP-PRINT () :PUBLIC? TRUE)" (CL:FUNCTION STARTUP-PRINT)
     NULL)
    (CL:LET* ((FUNCTION (LOOKUP-FUNCTION SYM-PRINT-LOGIC-STARTUP-PRINT)))
     (SET-DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS FUNCTION)
      SYM-PRINT-STELLA-METHOD-STARTUP-CLASSNAME
      (WRAP-STRING "Startup-Print") NULL-STRING-WRAPPER)))
   (CL:WHEN (CURRENT-STARTUP-TIME-PHASE? 8) (FINALIZE-SLOTS)
    (CLEANUP-UNFINALIZED-CLASSES))
   (CL:WHEN (CURRENT-STARTUP-TIME-PHASE? 9)
    (DEFINE-STELLA-GLOBAL-VARIABLE-FROM-STRINGIFIED-SOURCE
     "(DEFSPECIAL *PRINTMODE* KEYWORD :FLAT :DOCUMENTATION \"Controls the kind of detail that gets printed about
individual objects.  Values are :ORIGINAL, :REALISTIC, :FLAT, :DEBUG-LOW, and
 :DEBUG-HIGH.\")")
    (DEFINE-STELLA-GLOBAL-VARIABLE-FROM-STRINGIFIED-SOURCE
     "(DEFSPECIAL *PRINTFUNCTIONSASRELATIONS?* BOOLEAN FALSE :DOCUMENTATION \"Controls whether functions are printed as relations (F i v) or
as functions (= (F i) v).\")")
    (DEFINE-STELLA-GLOBAL-VARIABLE-FROM-STRINGIFIED-SOURCE
     "(DEFCONSTANT VARIABLE-BINDING-SEPARATOR STRING \"/\")")
    (DEFINE-STELLA-GLOBAL-VARIABLE-FROM-STRINGIFIED-SOURCE
     "(DEFCONSTANT VALUE-OF-SEPARATOR STRING \"//\")")
    (DEFINE-STELLA-GLOBAL-VARIABLE-FROM-STRINGIFIED-SOURCE
     "(DEFSPECIAL *SKOLEMNAMEMAPPINGTABLE* (KEY-VALUE-LIST OF SKOLEM SKOLEM) NULL :DOCUMENTATION \"If set, maps some skolems to names other than
their own.\")")
    (DEFINE-STELLA-GLOBAL-VARIABLE-FROM-STRINGIFIED-SOURCE
     "(DEFSPECIAL *PRINTINFRAME* CONTROL-FRAME NULL :DOCUMENTATION \"If set, controls diagnostic printing by making
variable bindings appear relative to the frame '*printInFrame*'.\")")
    (DEFINE-STELLA-GLOBAL-VARIABLE-FROM-STRINGIFIED-SOURCE
     "(DEFGLOBAL *DEBUG-MODE?* BOOLEAN FALSE :DOCUMENTATION \"If set to TRUE, simplify printing so that it
doesn't break during a debugging session.\")"))))
