;;; -*- Mode: Lisp; Package: STELLA; Syntax: COMMON-LISP; Base: 10 -*-

#|--------------------------------------------------------------------------+
 |                                                                          |
 |  COPYRIGHT (C) UNIVERSITY OF SOUTHERN CALIFORNIA, 1996-2000              | 
 |  University of Southern California, Information Sciences Institute       |
 |  4676 Admiralty Way                                                      |
 |  Marina Del Rey, California 90292                                        |
 |                                                                          |
 |  This software was developed under the terms and conditions of Contract  |
 |  No. N00014-94-C-0245 between the Defense Advanced Research Projects     |
 |  Agency and the University of Southern California, Information Sciences  | 
 |  Institute.  Use and distribution of this software is further subject    |
 |  to the provisions of that contract and any other agreements developed   |
 |  between the user of the software and the University of Southern         |
 |  California, Information Sciences Institute.  It is supplied "AS IS",    |
 |  without any warranties of any kind.  It is furnished only on the basis  |
 |  that any party who receives it indemnifies and holds harmless the       |
 |  parties who furnish and originate it against any claims, demands, or    |
 |  liabilities connected with using it, furnishing it to others or         |
 |  providing it to a third party.  THIS NOTICE MUST NOT BE REMOVED FROM    |
 |  THE SOFTWARE, AND IN THE EVENT THAT THE SOFTWARE IS DIVIDED, IT SHOULD  |
 |  BE ATTACHED TO EVERY PART.                                              |
 |                                                                          |
 +--------------------------------------------------------------------------|#

(CL:IN-PACKAGE "STELLA")

;;; Auxiliary variables:

(CL:DEFVAR SGT-JAVA-CLASS-OUT-STELLA-OBJECT NULL)
(CL:DEFVAR KWD-JAVA-CLASS-OUT-EMBEDDED NULL)
(CL:DEFVAR KWD-JAVA-CLASS-OUT-INSTANCE NULL)
(CL:DEFVAR KWD-JAVA-CLASS-OUT-CLASS NULL)
(CL:DEFVAR SYM-JAVA-CLASS-OUT-STELLA-METHOD NULL)
(CL:DEFVAR SYM-JAVA-CLASS-OUT-STELLA-MACRO NULL)
(CL:DEFVAR SYM-JAVA-CLASS-OUT-STELLA-PRINT-METHOD NULL)
(CL:DEFVAR SYM-JAVA-CLASS-OUT-STELLA-METHOD-CONSTRUCTOR? NULL)
(CL:DEFVAR SYM-JAVA-CLASS-OUT-STELLA-JAVA_EXCEPTION_CLASS NULL)
(CL:DEFVAR SYM-JAVA-CLASS-OUT-STELLA-JAVA_CLASS NULL)
(CL:DEFVAR SYM-JAVA-CLASS-OUT-STELLA-DOCUMENTATION NULL)
(CL:DEFVAR SYM-JAVA-CLASS-OUT-STELLA-JAVA_DECLARATIONS NULL)
(CL:DEFVAR SYM-JAVA-CLASS-OUT-STELLA-STARTUP-JAVA-CLASS-OUT NULL)
(CL:DEFVAR SYM-JAVA-CLASS-OUT-STELLA-METHOD-STARTUP-CLASSNAME NULL)

;;; Forward declarations:

(CL:DECLAIM
 (CL:SPECIAL *STELLA-MODULE* *MODULE* *CONTEXT* NULL-STRING-WRAPPER
  *TRANSLATIONUNITS* FALSE-WRAPPER))

;;; (DEFUN (JAVA-YIELD-INSTANCE-VAR-TREE CONS) ...)

(CL:DEFUN JAVA-YIELD-INSTANCE-VAR-TREE (SLOT CLASS)
  (CL:LET*
   ((SLOTNAME (TRUE-SLOT-NAME (%SLOT-NAME SLOT) (%CLASS-TYPE CLASS)))
    (SLOTTYPE
     (CL:IF (%SLOT-CONTEXT-SENSITIVE? SLOT)
      SGT-JAVA-CLASS-OUT-STELLA-OBJECT
      (COMPUTE-RETURN-TYPE-SPEC SLOT (%CLASS-TYPE CLASS))))
    (ALLOCATION (ALLOCATION SLOT)) (ALLOCATIONFORM NULL)
    (ACCESSFORM NULL))
   (CL:IF (%SLOT-PUBLIC? SLOT)
    (CL:SETQ ACCESSFORM (CONS (WRAP-STRING "public") NIL))
    (CL:SETQ ACCESSFORM (CONS (WRAP-STRING "protected") NIL)))
   (CL:COND
    ((CL:EQ ALLOCATION KWD-JAVA-CLASS-OUT-EMBEDDED)
     (CL:SETQ ALLOCATIONFORM NIL))
    ((CL:EQ ALLOCATION KWD-JAVA-CLASS-OUT-INSTANCE)
     (CL:SETQ ALLOCATIONFORM NIL))
    ((CL:EQ ALLOCATION KWD-JAVA-CLASS-OUT-CLASS)
     (CL:SETQ ALLOCATIONFORM (CONS (WRAP-STRING "static") NIL)))
    (CL:T (CL:SETQ ALLOCATIONFORM NIL)))
   (CL:RETURN-FROM JAVA-YIELD-INSTANCE-VAR-TREE
    (LIST*
     (CONCATENATE ACCESSFORM
      (CONCATENATE ALLOCATIONFORM
       (CONS (JAVA-TRANSLATE-TYPE-SPEC SLOTTYPE) NIL)))
     (JAVA-TRANSLATE-NAME SLOTNAME) NIL))))

;;; (DEFUN (JAVA-GET-INSTANCE-VARIABLE-DEFINITIONS CONS) ...)

(CL:DEFUN JAVA-GET-INSTANCE-VARIABLE-DEFINITIONS (CLASS)
  (CL:WHEN (%MIXIN? CLASS)
   (CL:RETURN-FROM JAVA-GET-INSTANCE-VARIABLE-DEFINITIONS NIL))
  (CL:LET* ((INSTANCEVARDEFS NIL) (SUPERCLASS NULL))
   (CL:LET*
    ((SLOT NULL) (ITER-000 (CLASS-SLOTS CLASS)) (COLLECT-000 NULL))
    (CL:LOOP WHILE (NEXT? ITER-000) DO (CL:SETQ SLOT (%VALUE ITER-000))
     (CL:WHEN
      (CL:AND (STORAGE-SLOT? SLOT) (NATIVE-SLOT? SLOT)
       (CL:EQ (NATIVE-SLOT-HOME SLOT CLASS) CLASS))
      (CL:IF (CL:EQ COLLECT-000 NULL)
       (CL:PROGN
        (CL:SETQ COLLECT-000
         (CONS (JAVA-YIELD-INSTANCE-VAR-TREE SLOT CLASS) NIL))
        (CL:IF (CL:EQ INSTANCEVARDEFS NIL)
         (CL:SETQ INSTANCEVARDEFS COLLECT-000)
         (ADD-CONS-TO-END-OF-CONS-LIST INSTANCEVARDEFS COLLECT-000)))
       (CL:PROGN
        (CL:SETF (%%REST COLLECT-000)
         (CONS (JAVA-YIELD-INSTANCE-VAR-TREE SLOT CLASS) NIL))
        (CL:SETQ COLLECT-000 (%%REST COLLECT-000)))))))
   (CL:LET*
    ((SUPER NULL)
     (ITER-001 (%THE-CONS-LIST (%CLASS-DIRECT-SUPERS CLASS))))
    (CL:LOOP WHILE (CL:NOT (CL:EQ ITER-001 NIL)) DO
     (CL:PROGN (CL:SETQ SUPER (%%VALUE ITER-001))
      (CL:SETQ ITER-001 (%%REST ITER-001)))
     (CL:SETQ SUPERCLASS (GET-STELLA-CLASS SUPER TRUE))
     (CL:WHEN (%MIXIN? SUPERCLASS)
      (CL:LET*
       ((SLOT NULL) (ITER-002 (CLASS-SLOTS SUPERCLASS))
        (COLLECT-001 NULL))
       (CL:LOOP WHILE (NEXT? ITER-002) DO (CL:SETQ SLOT (%VALUE ITER-002))
        (CL:WHEN
         (CL:AND (STORAGE-SLOT? SLOT) (NATIVE-SLOT? SLOT)
          (CL:EQ (NATIVE-SLOT-HOME SLOT SUPERCLASS) SUPERCLASS))
         (CL:IF (CL:EQ COLLECT-001 NULL)
          (CL:PROGN
           (CL:SETQ COLLECT-001
            (CONS (JAVA-YIELD-INSTANCE-VAR-TREE SLOT SUPERCLASS) NIL))
           (CL:IF (CL:EQ INSTANCEVARDEFS NIL)
            (CL:SETQ INSTANCEVARDEFS COLLECT-001)
            (ADD-CONS-TO-END-OF-CONS-LIST INSTANCEVARDEFS COLLECT-001)))
          (CL:PROGN
           (CL:SETF (%%REST COLLECT-001)
            (CONS (JAVA-YIELD-INSTANCE-VAR-TREE SLOT SUPERCLASS) NIL))
           (CL:SETQ COLLECT-001 (%%REST COLLECT-001))))))))))
   (CL:RETURN-FROM JAVA-GET-INSTANCE-VARIABLE-DEFINITIONS INSTANCEVARDEFS)))

;;; (DEFUN (JAVA-GET-CONSTRUCTOR-AND-METHOD-DEFINITIONS CONS CONS) ...)

(CL:DEFUN JAVA-GET-CONSTRUCTOR-AND-METHOD-DEFINITIONS (CLASS)
  (CL:LET*
   ((CONSTRUCTORDEFINITIONS NIL) (METHODDEFINITIONS NIL) (METHOD NULL))
   (CL:LET* ((TU NULL) (ITER-000 (%THE-CONS-LIST *TRANSLATIONUNITS*)))
    (CL:LOOP WHILE (CL:NOT (CL:EQ ITER-000 NIL)) DO
     (CL:PROGN (CL:SETQ TU (%%VALUE ITER-000))
      (CL:SETQ ITER-000 (%%REST ITER-000)))
     (CL:WHEN
      (CL:OR (CL:EQ (%CATEGORY TU) SYM-JAVA-CLASS-OUT-STELLA-METHOD)
       (CL:EQ (%CATEGORY TU) SYM-JAVA-CLASS-OUT-STELLA-MACRO)
       (CL:EQ (%CATEGORY TU) SYM-JAVA-CLASS-OUT-STELLA-PRINT-METHOD))
      (CL:IF (CL:EQ (%CATEGORY TU) SYM-JAVA-CLASS-OUT-STELLA-PRINT-METHOD)
       (CL:WHEN
        (CL:EQ (CANONICAL-TYPE (%CLASS-TYPE (%THE-OBJECT TU)))
         (%CLASS-TYPE CLASS))
        (CL:SETQ METHODDEFINITIONS (CONS TU METHODDEFINITIONS)))
       (CL:PROGN (CL:SETQ METHOD (%THE-OBJECT TU))
        (CL:COND
         ((CL:OR (JAVA-FLOTSAM-FUNCTION? METHOD)
           (METHOD-STARTUP-FUNCTION? METHOD)))
         ((%WRAPPER-VALUE
           (DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS METHOD)
            SYM-JAVA-CLASS-OUT-STELLA-METHOD-CONSTRUCTOR? FALSE-WRAPPER))
          (CL:WHEN
           (CL:EQ (CANONICAL-TYPE (%SLOT-BASE-TYPE METHOD))
            (%CLASS-TYPE CLASS))
           (CL:SETQ CONSTRUCTORDEFINITIONS
            (CONS TU CONSTRUCTORDEFINITIONS))))
         ((CL:EQ (CANONICAL-TYPE (%SLOT-OWNER METHOD))
           (%CLASS-TYPE CLASS))
          (CL:SETQ METHODDEFINITIONS (CONS TU METHODDEFINITIONS)))))))))
   (CL:RETURN-FROM JAVA-GET-CONSTRUCTOR-AND-METHOD-DEFINITIONS
    (CL:VALUES (REVERSE CONSTRUCTORDEFINITIONS)
     (REVERSE METHODDEFINITIONS)))))

;;; (DEFUN (JAVA-TRANSLATE-DEFINE-NATIVE-CLASS-UNIT CONS) ...)

(CL:DEFUN JAVA-TRANSLATE-DEFINE-NATIVE-CLASS-UNIT (UNIT)
  (CL:RETURN-FROM JAVA-TRANSLATE-DEFINE-NATIVE-CLASS-UNIT
   (JAVA-TRANSLATE-DEFINE-NATIVE-CLASS (%THE-OBJECT UNIT))))

;;; (DEFUN (JAVA-TRANSLATE-DEFINE-NATIVE-CLASS CONS) ...)

(CL:DEFUN JAVA-TRANSLATE-DEFINE-NATIVE-CLASS (CLASS)
  (CL:LET*
   ((PARENTCLASS NULL) (INTERFACECLASSES NIL) (INSTANCEVARDEFS NIL)
    (CONSTRUCTORDEFS NIL) (MODIFIERS NIL) (METHODDEFS NIL))
   (CL:LET*
    ((*CONTEXT* (HOME-MODULE CLASS)) (*MODULE* (%BASE-MODULE *CONTEXT*)))
    (CL:DECLARE (CL:SPECIAL *CONTEXT* *MODULE*))
    (CL:WHEN (%MIXIN? CLASS)
     (CL:RETURN-FROM JAVA-TRANSLATE-DEFINE-NATIVE-CLASS NIL))
    (CL:WHEN (%ABSTRACT? CLASS)
     (CL:SETQ MODIFIERS (CONS (WRAP-STRING "abstract") MODIFIERS)))
    (CL:WHEN (PUBLIC? CLASS)
     (CL:SETQ MODIFIERS (CONS (WRAP-STRING "public") MODIFIERS)))
    (CL:WHEN (CL:NOT (CL:EQ (FIRST (%CLASS-DIRECT-SUPERS CLASS)) NULL))
     (CL:SETQ PARENTCLASS
      (JAVA-TRANSLATE-CLASS-NAME
       (SYMBOLIZE (FIRST (%CLASS-DIRECT-SUPERS CLASS))))))
    (CL:SETQ INSTANCEVARDEFS
     (JAVA-GET-INSTANCE-VARIABLE-DEFINITIONS CLASS))
    (CL:MULTIPLE-VALUE-SETQ (CONSTRUCTORDEFS METHODDEFS)
     (JAVA-GET-CONSTRUCTOR-AND-METHOD-DEFINITIONS CLASS))
    (CL:RETURN-FROM JAVA-TRANSLATE-DEFINE-NATIVE-CLASS
     (CONS
      (CL:IF (EXCEPTION-CLASS? CLASS)
       SYM-JAVA-CLASS-OUT-STELLA-JAVA_EXCEPTION_CLASS
       SYM-JAVA-CLASS-OUT-STELLA-JAVA_CLASS)
      (CONS
       (CL:IF
        (CL:NOT
         (CL:EQ
          (%WRAPPER-VALUE
           (DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS CLASS)
            SYM-JAVA-CLASS-OUT-STELLA-DOCUMENTATION NULL-STRING-WRAPPER))
          STELLA::NULL-STRING))
        (WRAP-STRING
         (%WRAPPER-VALUE
          (DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS CLASS)
           SYM-JAVA-CLASS-OUT-STELLA-DOCUMENTATION NULL-STRING-WRAPPER)))
        NULL)
       (CONS MODIFIERS
        (CONS
         (CL:IF (%MIXIN? CLASS) (WRAP-STRING "interface")
          (WRAP-STRING "class"))
         (CONS (JAVA-TRANSLATE-CLASS-NAME (SYMBOLIZE (%CLASS-TYPE CLASS)))
          (CONS PARENTCLASS
           (CONS INTERFACECLASSES
            (CONS
             (CL:IF (CL:NOT (CL:EQ INSTANCEVARDEFS NIL))
              (CONS
               (CONS SYM-JAVA-CLASS-OUT-STELLA-JAVA_DECLARATIONS
                (CONCATENATE INSTANCEVARDEFS NIL))
               NIL)
              NIL)
             (CONS CONSTRUCTORDEFS (CONS METHODDEFS NIL))))))))))))))

(CL:DEFUN STARTUP-JAVA-CLASS-OUT ()
  (CL:LET* ((*MODULE* *STELLA-MODULE*) (*CONTEXT* *MODULE*))
   (CL:DECLARE (CL:SPECIAL *MODULE* *CONTEXT*))
   (CL:WHEN (CURRENT-STARTUP-TIME-PHASE? 2)
    (CL:SETQ SGT-JAVA-CLASS-OUT-STELLA-OBJECT
     (INTERN-RIGID-SYMBOL-WRT-MODULE "OBJECT" NULL 1))
    (CL:SETQ KWD-JAVA-CLASS-OUT-EMBEDDED
     (INTERN-RIGID-SYMBOL-WRT-MODULE "EMBEDDED" NULL 2))
    (CL:SETQ KWD-JAVA-CLASS-OUT-INSTANCE
     (INTERN-RIGID-SYMBOL-WRT-MODULE "INSTANCE" NULL 2))
    (CL:SETQ KWD-JAVA-CLASS-OUT-CLASS
     (INTERN-RIGID-SYMBOL-WRT-MODULE "CLASS" NULL 2))
    (CL:SETQ SYM-JAVA-CLASS-OUT-STELLA-METHOD
     (INTERN-RIGID-SYMBOL-WRT-MODULE "METHOD" NULL 0))
    (CL:SETQ SYM-JAVA-CLASS-OUT-STELLA-MACRO
     (INTERN-RIGID-SYMBOL-WRT-MODULE "MACRO" NULL 0))
    (CL:SETQ SYM-JAVA-CLASS-OUT-STELLA-PRINT-METHOD
     (INTERN-RIGID-SYMBOL-WRT-MODULE "PRINT-METHOD" NULL 0))
    (CL:SETQ SYM-JAVA-CLASS-OUT-STELLA-METHOD-CONSTRUCTOR?
     (INTERN-RIGID-SYMBOL-WRT-MODULE "METHOD-CONSTRUCTOR?" NULL 0))
    (CL:SETQ SYM-JAVA-CLASS-OUT-STELLA-JAVA_EXCEPTION_CLASS
     (INTERN-RIGID-SYMBOL-WRT-MODULE "JAVA_EXCEPTION_CLASS" NULL 0))
    (CL:SETQ SYM-JAVA-CLASS-OUT-STELLA-JAVA_CLASS
     (INTERN-RIGID-SYMBOL-WRT-MODULE "JAVA_CLASS" NULL 0))
    (CL:SETQ SYM-JAVA-CLASS-OUT-STELLA-DOCUMENTATION
     (INTERN-RIGID-SYMBOL-WRT-MODULE "DOCUMENTATION" NULL 0))
    (CL:SETQ SYM-JAVA-CLASS-OUT-STELLA-JAVA_DECLARATIONS
     (INTERN-RIGID-SYMBOL-WRT-MODULE "JAVA_DECLARATIONS" NULL 0))
    (CL:SETQ SYM-JAVA-CLASS-OUT-STELLA-STARTUP-JAVA-CLASS-OUT
     (INTERN-RIGID-SYMBOL-WRT-MODULE "STARTUP-JAVA-CLASS-OUT" NULL 0))
    (CL:SETQ SYM-JAVA-CLASS-OUT-STELLA-METHOD-STARTUP-CLASSNAME
     (INTERN-RIGID-SYMBOL-WRT-MODULE "METHOD-STARTUP-CLASSNAME" NULL 0)))
   (CL:WHEN (CURRENT-STARTUP-TIME-PHASE? 6) (FINALIZE-CLASSES))
   (CL:WHEN (CURRENT-STARTUP-TIME-PHASE? 7)
    (DEFINE-FUNCTION-OBJECT "JAVA-YIELD-INSTANCE-VAR-TREE"
     "(DEFUN (JAVA-YIELD-INSTANCE-VAR-TREE CONS) ((SLOT STORAGE-SLOT) (CLASS CLASS)))"
     (CL:FUNCTION JAVA-YIELD-INSTANCE-VAR-TREE) NULL)
    (DEFINE-FUNCTION-OBJECT "JAVA-GET-INSTANCE-VARIABLE-DEFINITIONS"
     "(DEFUN (JAVA-GET-INSTANCE-VARIABLE-DEFINITIONS CONS) ((CLASS CLASS)))"
     (CL:FUNCTION JAVA-GET-INSTANCE-VARIABLE-DEFINITIONS) NULL)
    (DEFINE-FUNCTION-OBJECT "JAVA-GET-CONSTRUCTOR-AND-METHOD-DEFINITIONS"
     "(DEFUN (JAVA-GET-CONSTRUCTOR-AND-METHOD-DEFINITIONS CONS CONS) ((CLASS CLASS)))"
     (CL:FUNCTION JAVA-GET-CONSTRUCTOR-AND-METHOD-DEFINITIONS) NULL)
    (DEFINE-FUNCTION-OBJECT "JAVA-TRANSLATE-DEFINE-NATIVE-CLASS-UNIT"
     "(DEFUN (JAVA-TRANSLATE-DEFINE-NATIVE-CLASS-UNIT CONS) ((UNIT TRANSLATION-UNIT)))"
     (CL:FUNCTION JAVA-TRANSLATE-DEFINE-NATIVE-CLASS-UNIT) NULL)
    (DEFINE-FUNCTION-OBJECT "JAVA-TRANSLATE-DEFINE-NATIVE-CLASS"
     "(DEFUN (JAVA-TRANSLATE-DEFINE-NATIVE-CLASS CONS) ((CLASS CLASS)))"
     (CL:FUNCTION JAVA-TRANSLATE-DEFINE-NATIVE-CLASS) NULL)
    (DEFINE-FUNCTION-OBJECT "STARTUP-JAVA-CLASS-OUT"
     "(DEFUN STARTUP-JAVA-CLASS-OUT () :PUBLIC? TRUE)"
     (CL:FUNCTION STARTUP-JAVA-CLASS-OUT) NULL)
    (CL:LET*
     ((FUNCTION
       (LOOKUP-FUNCTION
        SYM-JAVA-CLASS-OUT-STELLA-STARTUP-JAVA-CLASS-OUT)))
     (SET-DYNAMIC-SLOT-VALUE (%DYNAMIC-SLOTS FUNCTION)
      SYM-JAVA-CLASS-OUT-STELLA-METHOD-STARTUP-CLASSNAME
      (WRAP-STRING "Startup-Java-Class-Out") NULL-STRING-WRAPPER)))
   (CL:WHEN (CURRENT-STARTUP-TIME-PHASE? 8) (FINALIZE-SLOTS)
    (CLEANUP-UNFINALIZED-CLASSES)))
  :VOID)
